#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <sys/time.h>
#include <GL/glut.h>
#include "vmath/vmath.h"

void redraw(void);

quat_t get_quat(float t);
void gl_rotate_quat(quat_t q);

void draw_teapot(float time);
void draw_grid(float spacing, float sz);
void draw_graph(float time);

void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mouse_button(int bn, int state, int x, int y);
void mouse_motion(int x, int y);

#ifdef FREEGLUT
#include <GL/freeglut_ext.h>
#define MOUSEWHEEL
void mouse_wheel(int unk, int dir, int x, int y);
#endif

quat_t *load_keys(const char *fname, int *count);
unsigned int get_msec(void);

float cam_x, cam_y, cam_z = 6;
float cam_rot, cam_pitch = 34;
float pan_offs = 0.05;

int xsz, ysz;

quat_t *keys;
int key_count;

float anm_time;
int run_time;


int main(int argc, char **argv)
{
	float amb[] = {0, 0, 0, 0};

	if(!(keys = load_keys("quat_keys", &key_count))) {
		return EXIT_FAILURE;
	}

	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("foo");
	xsz = glutGet(GLUT_WINDOW_WIDTH);
	ysz = glutGet(GLUT_WINDOW_HEIGHT);

	glutDisplayFunc(redraw);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse_button);
	glutMotionFunc(mouse_motion);
	glutReshapeFunc(reshape);
#ifdef MOUSEWHEEL
	glutMouseWheelFunc(mouse_wheel);
#endif

	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0, (float)xsz / (float)ysz, 1.0, 1000.0);

	glutMainLoop();
	return 0;
}

void redraw(void)
{
	float ltpos[] = {-50, 60, 100, 1};

	if(run_time) {
		anm_time = fmod(get_msec() / 10000.0, 1.0);
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(-cam_x, -cam_y, -cam_z);
	glRotatef(cam_pitch, 1, 0, 0);
	glRotatef(cam_rot, 0, 1, 0);

	glLightfv(GL_LIGHT0, GL_POSITION, ltpos);

	draw_grid(2.0, 100.0);

	draw_teapot(anm_time);

	glutSwapBuffers();
}

quat_t get_quat(float t)
{
	float kpos = t * key_count;
	int a = floor(kpos);
	int b = ceil(kpos);
	
	printf("a(%d): [%.3f  %.3f  %.3f]  %.3f\n", a, keys[a].x, keys[a].y, keys[a].z, keys[a].w);
	printf("b(%d): [%.3f  %.3f  %.3f]  %.3f\n", b, keys[b].x, keys[b].y, keys[b].z, keys[b].w);

	return a == b ? keys[a] : quat_slerp(keys[a], keys[b], kpos - a);
}

void gl_rotate_quat(quat_t q)
{
	float glm[16];
	mat3_t m;

	quat_to_mat3(m, q);

	glm[0] = m[0][0]; glm[1] = m[0][1]; glm[2] = m[0][2]; glm[3] = 0;
	glm[4] = m[1][0]; glm[5] = m[1][1]; glm[6] = m[1][2]; glm[7] = 0;
	glm[8] = m[2][0]; glm[9] = m[2][1]; glm[10] = m[2][2]; glm[11] = 0;
	glm[12] = glm[13] = glm[14] = 0; glm[15] = 1;

	glMultMatrixf(glm);
}

void draw_teapot(float time)
{
	float col[] = {1.0, 0.7, 0.4, 1.0};
	float spec[] = {1, 1, 1, 1};

	glTranslatef(0, 0.7, 0);
	gl_rotate_quat(get_quat(time));

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 60.0f);

	glFrontFace(GL_CW);
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);
}

void draw_grid(float spacing, float sz)
{
	float hsz = sz * 0.5;
	int i, lt_enabl, lines = (int)(hsz / spacing) + 1;

	lt_enabl = glIsEnabled(GL_LIGHTING);
	glDisable(GL_LIGHTING);

	glBegin(GL_LINES);
	glColor3f(0.4, 0.4, 0.4);
	for(i=1; i<lines; i++) {
		glVertex3f(i * spacing, 0, -hsz);
		glVertex3f(i * spacing, 0, hsz);
		glVertex3f(-i * spacing, 0, -hsz);
		glVertex3f(-i * spacing, 0, hsz);
		glVertex3f(-hsz, 0, i * spacing);
		glVertex3f(hsz, 0, i * spacing);
		glVertex3f(-hsz, 0, -i * spacing);
		glVertex3f(hsz, 0, -i * spacing);
	}
	glEnd();

	glLineWidth(2.0);
	glBegin(GL_LINES);
	glColor3f(0.5, 0.5, 0.5);
	glVertex3f(-hsz, 0, 0);
	glVertex3f(hsz, 0, 0);
	glVertex3f(0, 0, -hsz);
	glVertex3f(0, 0, hsz);
	glEnd();
	glLineWidth(1.0);

	if(lt_enabl) {
		glEnable(GL_LIGHTING);
	}
}


/* ---- input handlers ---- */

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	xsz = x;
	ysz = y;
}


static int prev_x = -1;
static int prev_y = -1;
static int pbn;
static int ctrl;

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
	case 'q':
	case 'Q':
		exit(0);

	case ' ':
		run_time = !run_time;
		glutIdleFunc(run_time ? glutPostRedisplay : 0);
		break;

	default:
		break;
	}
}

void mouse_button(int button, int state, int x, int y)
{
	if(state == GLUT_DOWN) {
		prev_x = x;
		prev_y = y;
		pbn = button;
		ctrl = (glutGetModifiers() & GLUT_ACTIVE_CTRL) ? 1 : 0;
	} else {
		prev_x = -1;
		prev_y = -1;
	}
}


void mouse_motion(int x, int y)
{
	if(pbn == GLUT_LEFT_BUTTON) {
		if(ctrl) {
			anm_time = fmod(anm_time + (float)(x - prev_x) / 600.0f, 1.0);
			while(anm_time < 0.0) {
				anm_time += 1.0;
			}
		} else {
			cam_x += pan_offs * (float)(prev_x - x) / 2.0f;
			cam_y += pan_offs * (float)(y - prev_y) / 2.0f;
		}

		prev_x = x;
		prev_y = y;
		glutPostRedisplay();

	} else if(pbn == GLUT_MIDDLE_BUTTON) {
		cam_rot += (float)(x - prev_x) / 2.0f;
		cam_pitch += (float)(y - prev_y) / 2.0f;
		if(cam_pitch > 90.0) cam_pitch = 90.0;
		if(cam_pitch < -90.0) cam_pitch = -90.0;
		
		prev_x = x;
		prev_y = y;
		glutPostRedisplay();

	} else if(pbn == GLUT_RIGHT_BUTTON) {
		cam_z -= (float)(prev_y - y) / 10.0f;
		prev_y = y;
		glutPostRedisplay();
	}

}

#ifdef MOUSEWHEEL
void mouse_wheel(int unk, int dir, int x, int y)
{
	if(dir > 0) {
		cam_z -= pan_offs * 10.0;
	} else {
		cam_z += pan_offs * 10.0;
	}
	glutPostRedisplay();
}
#endif

quat_t *load_keys(const char *fname, int *count)
{
	quat_t *keys = 0;
	FILE *fp;
	int i, knum;
	char buf[512];

	if(!(fp = fopen(fname, "r"))) {
		sprintf(buf, "failed to open file `%s'", fname);
		perror(buf);
		return 0;
	}

	knum = 0;
	while(fgets(buf, sizeof buf, fp)) {
		knum++;
	}
	fseek(fp, 0, SEEK_SET);

	if(!(keys = malloc(knum * sizeof *keys))) {
		perror("malloc failed");
		fclose(fp);
		return 0;
	}

	i = 0;
	while(fgets(buf, sizeof buf, fp)) {
		char *ptr;

		if(!isdigit(*buf) && *buf != '-' && *buf != '.') {
			goto corrupt;
		}
		keys[i].x = atof(buf);

		if(!(ptr = strchr(buf, ' '))) {
			goto corrupt;
		}
		if(!isdigit(*++ptr) && *ptr != '-' && *ptr != '.') {
			goto corrupt;
		}
		keys[i].y = atof(ptr);

		if(!(ptr = strchr(ptr, ' '))) {
			goto corrupt;
		}
		if(!isdigit(*++ptr) && *ptr != '-' && *ptr != '.') {
			goto corrupt;
		}
		keys[i].z = atof(ptr);

		if(!(ptr = strchr(ptr, ' '))) {
			goto corrupt;
		}
		if(!isdigit(*++ptr) && *ptr != '-' && *ptr != '.') {
			goto corrupt;
		}
		keys[i].w = atof(ptr);

		keys[i] = quat_normalize(keys[i]);

		i++;
	}
	assert(i == knum);

	fclose(fp);
	if(count) *count = knum - 1;
	return keys;

corrupt:
	free(keys);
	fclose(fp);
	fprintf(stderr, "keyframe file corrupt or invalid\n");
	return 0;
}

unsigned int get_msec(void)
{
	static struct timeval timeval, first_timeval;
	
	gettimeofday(&timeval, 0);

	if(first_timeval.tv_sec == 0) {
		first_timeval = timeval;
		return 0;
	}
	return (timeval.tv_sec - first_timeval.tv_sec) * 1000 + (timeval.tv_usec - first_timeval.tv_usec) / 1000;
}
