#include <stdio.h>
#include <math.h>
#include "quat.h"

void quat_print(FILE *fp, quat_t q)
{
	fprintf(fp, "([ %.4f %.4f %.4f ] %.4f)", q.x, q.y, q.z, q.w);
}

quat_t quat_rotate(quat_t q, scalar_t angle, scalar_t x, scalar_t y, scalar_t z)
{
	quat_t rq;
	scalar_t half_angle = angle * 0.5;
	scalar_t sin_half = sin(half_angle);

	rq.w = cos(half_angle);
	rq.x = x * sin_half;
	rq.y = y * sin_half;
	rq.z = z * sin_half;

	return quat_mul(q, rq);
}

quat_t quat_rotate_quat(quat_t q, quat_t rotq)
{
	return quat_mul(quat_mul(rotq, q), quat_conjugate(rotq));
}

#include <float.h>

quat_t quat_slerp(quat_t q1, quat_t q2, scalar_t t)
{
	quat_t res;
	scalar_t dot, angle, sin_angle, s1, s2;

	dot = q1.w * q2.w + q1.x * q2.x + q1.y * q2.y + q1.z * q2.z;
	if(fabs(dot) < XSMALL_NUMBER) {
		return q1;
	}
	if(dot < 0.0) {
		q1.x = -q1.x;
		q1.y = -q1.y;
		q1.z = -q1.z;
		q1.w = -q1.w;
		dot = -dot;
	}
	
	if((1.0 - dot) >= 0.05) {
		angle = acos(dot);
		s1 = sin(angle * (1.0 - t)) / sin(angle);
		s2 = sin(angle * t) / sin(angle);
	} else {
		s1 = 1.0 - t;
		s2 = t;
	}

	res.x = q1.x * s1 + q2.x * s2;
	res.y = q1.y * s1 + q2.y * s2;
	res.z = q1.z * s1 + q2.z * s2;
	res.w = q1.w * s1 + q2.w * s2;
	return res;
}
